﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace ThreadWatchDogCon
{
    public class App
    {
        private static readonly double ProcessorFactor = Convert.ToDouble(Environment.ProcessorCount);
        private static readonly double ProcessThreshold = 1D / App.ProcessorFactor;
        private static readonly double ThreadThreshold = App.ProcessThreshold / 2D;

        private DateTime dt;

        private ProcessMonitor processMonitor = new ProcessMonitor();

        private string filter;

        private bool enable;

        public App()
        {
        }

        public App(string filter, bool enable)
        {
            this.filter = filter;
            this.enable = enable;
        }

        public void Execute()
        {
            Console.WriteLine("Process Monitor started.");
            Console.WriteLine("Processor factor: " + App.ProcessorFactor);
            Console.WriteLine("Process threshold: " + App.ProcessThreshold);
            Console.WriteLine("Thread threshold: " + App.ThreadThreshold);
            Console.WriteLine("Process min history: " + ProcessMonitor.ProcessMinHistory);
            Console.WriteLine("Process max history: " + ProcessMonitor.ProcessMaxHistory);
            Console.WriteLine("Thread min history: " + ProcessMonitor.ThreadMinHistory);
            Console.WriteLine("Thread max history: " + ProcessMonitor.ThreadMaxHistory);

            var t = new Timer(this.TimerCallback, null, 0, 1000);
            Console.ReadLine();
            t.Dispose();

            Console.WriteLine("Process Monitor stopped.");
        }

        private void TimerCallback(object state)
        {
            dt = DateTime.Now;

            lock (processMonitor)
            {
                this.processMonitor.Refresh();

                foreach (var pi in processMonitor.Processes)
                {
                    this.OutputProcess(pi);
                }
            }
        }

        private void OutputProcess(ProcessInfo pi)
        {
            try
            {
                if ((pi.MainModuleFileName ?? string.Empty).IndexOf((this.filter ?? string.Empty), StringComparison.OrdinalIgnoreCase) >= 0
                    && pi.ProcessorTimeHistory.LongCount() >= ProcessMonitor.ProcessMinHistory)
                {
                    var last = pi.ProcessorTimeHistory.Last();
                    var first = pi.ProcessorTimeHistory.First();

                    var pct = this.CalcPct(first, last);

                    if (pct > App.ProcessThreshold)
                    {
                        string s = dt.ToString("HH':'mm':'ss")
                            + "  "
                            + (pct * 100D).ToString("N3") + "%"
                            + "  "
                            + "P"
                            + "  "
                            + string.Format("{0,5}", pi.Id)
                            + "  "
                            + pi.Process.ProcessName
                            + "  "
                            + pi.MainModuleFileName;
                        Console.WriteLine(s);

                        foreach (var ti in pi.Threads)
                        {
                            this.OutputThread(ti);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("OutputProcess: " + pi.Id + " " + ex.Message);
            }
        }

        private void OutputThread(ThreadInfo ti)
        {
            try
            {
                if (ti.ProcessorTimeHistory.LongCount() >= ProcessMonitor.ThreadMinHistory)
                {
                    var last = ti.ProcessorTimeHistory.Last();
                    var first = ti.ProcessorTimeHistory.First();

                    var pct = this.CalcPct(first, last);

                    if (pct > App.ThreadThreshold)
                    {
                        string s = dt.ToString("HH':'mm':'ss")
                            + "  "
                            + (pct * 100D).ToString("N3") + "%"
                            + "  "
                            + "T"
                            + "  "
                            + string.Format("{0,5}", ti.Id);
                        Console.WriteLine(s);

                        bool isSuspended = (ti.Thread.ThreadState == System.Diagnostics.ThreadState.Wait
                            && ti.Thread.WaitReason == System.Diagnostics.ThreadWaitReason.Suspended);

                        if (!isSuspended)
                        {
                            if (this.enable)
                            {
                                Console.WriteLine("Suspending thread: " + ti.Thread.Id);
                                NativeMethods.SuspendThread(ti.Thread.Id);
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("OutputThread: " + ti.Id + " " + ex.Message);
            }
        }

        private double CalcPct(ProcessorTimeInfo x, ProcessorTimeInfo y)
        {
            double diff = 0D;
            double dur = 0D;
            double pct = 0D;

            if (y != null && x != null)
            {
                dur = Convert.ToDouble(y.Version - x.Version);
                if (dur > 0)
                {
                    diff = Convert.ToDouble((y.TotalProcessorTime - x.TotalProcessorTime).Ticks);
                    pct = diff / dur;

                    if (App.ProcessorFactor > 0)
                    {
                        return pct / App.ProcessorFactor;
                    }

                    return pct;
                }
            }

            return 0D;
        }
    }
}
