﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;

namespace ThreadWatchDogCon
{
    public class ProcessMonitor
    {
        // Use min for warmup
        internal const long ProcessMinHistory = 30L;
        internal const long ProcessMaxHistory = 90L;
        internal const long ThreadMinHistory = 30L;
        internal const long ThreadMaxHistory = 90L;

        // Process Id 0 - Idle
        // Process Id 4 - System
        private IEnumerable<int> processIdBlacklist = new int[] { 0, 4 };

        private IList<ProcessInfo> processes = new List<ProcessInfo>();

        private long version;

        public IList<ProcessInfo> Processes
        {
            get
            {
                return this.processes;
            }

            internal set
            {
                this.processes = value;
            }
        }

        public void Refresh()
        {
            this.version = DateTime.UtcNow.Ticks;

            this.RefreshProcesses();
        }

        private void RefreshProcesses()
        {
            int processId;

            foreach (Process p in Process.GetProcesses())
            {
                processId = p.Id;

                if (this.processIdBlacklist.Any(id => id == processId))
                {
                    continue;
                }

                try
                {
                    var tempPI = new ProcessInfo(p);
                    var pi2 = this.Processes.Where(pi => pi.Id == tempPI.Id && pi.StartTime == tempPI.StartTime).FirstOrDefault();

                    if (pi2 == null)
                    {
                        pi2 = tempPI;
                        this.Processes.Add(pi2);
                    }

                    // Always update Process
                    pi2.Process = p;

                    this.UpdateProcessHistory(pi2);
                    this.RefreshThreads(pi2);

                    pi2.Version = this.version;
                }
                catch (Exception ex)
                {
                    Console.WriteLine("RefreshProcesses: " + processId + " " + ex.Message);
                }
            }

            // Remove old processes from collection
            this.Processes = this.Processes.Where(a => a.Version == this.version).ToList();
        }

        private void UpdateProcessHistory(ProcessInfo pi)
        {
            try
            {
                if (!pi.Ignore)
                {
                    while (pi.ProcessorTimeHistory.LongCount() >= ProcessMonitor.ProcessMaxHistory)
                    {
                        pi.ProcessorTimeHistory.RemoveAt(0);
                    }

                    pi.ProcessorTimeHistory.Add(new ProcessorTimeInfo { TotalProcessorTime = pi.Process.TotalProcessorTime, Version = this.version });
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("UpdateProcessHistory: " + pi.Id + " " + ex.Message);
                pi.Ignore = true;
            }
        }

        private void RefreshThreads(ProcessInfo pi)
        {
            int threadId;

            foreach (ProcessThread pt in pi.Process.Threads)
            {
                threadId = pt.Id;

                try
                {
                    var tempTI = new ThreadInfo(pt);
                    var ti2 = pi.Threads.Where(ti => ti.Id == tempTI.Id && ti.StartTime == tempTI.StartTime).FirstOrDefault();

                    if (ti2 == null)
                    {
                        ti2 = tempTI;
                        pi.Threads.Add(ti2);
                    }

                    // Always update Thread
                    ti2.Thread = pt;

                    this.UpdateThreadHistory(ti2);

                    ti2.Version = this.version;
                }
                catch (Exception ex)
                {
                    Console.WriteLine("RefreshThreads: " + threadId + " " + ex.Message);
                }
            }

            // Remove old processes from collection
            pi.Threads = pi.Threads.Where(a => a.Version == this.version).ToList();
        }

        private void UpdateThreadHistory(ThreadInfo ti)
        {
            try
            {
                if (!ti.Ignore)
                {
                    while (ti.ProcessorTimeHistory.LongCount() >= ProcessMonitor.ThreadMaxHistory)
                    {
                        ti.ProcessorTimeHistory.RemoveAt(0);
                    }

                    ti.ProcessorTimeHistory.Add(new ProcessorTimeInfo { TotalProcessorTime = ti.Thread.TotalProcessorTime, Version = this.version });
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("UpdateThreadHistory: " + ti.Id + " " + ex.Message);
                ti.Ignore = true;
            }
        }
    }
}
